#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2012, Nathan Sheldon. All rights reserved.
# http://www.nathansheldon.com/files/Announcer-Plugin.php
#
#	Version 1.1.2
#
#	History:	1.1.2 (26-Jul-2012)
#				* Modified voice list code so 3rd party voices show up properly.
#				--
#				1.1.1 (06-Jul-2012)
#				* Fixed a bug that would prevent spoken messages from playing
#				  properly when an Announcer action was used.
#				--
#				1.1 (04-Jul-2012)
#				* Added the "Announce a Message" action capability.
#				* Changed the label for the "Speech Text Variable" to "Speech
#				  Message Variable" for consistency.
#				--
#				1.0 (02-Jul-2012)
#				* Initial release.
#

################################################################################
import time
import random
import os
import mactypes
import subprocess
from appscript import *
from AppKit import NSSpeechSynthesizer	# Used to get list of installed voices.

################################################################################
# Globals
################################################################################
# (none for now).

########################################
def cleanText(inString=""):
	# (This method isn't currently being used).
	inString = inString.replace("\\","")	# Remove possible escape attempts.
	inString = inString.replace("!",r"\!")	# Escape exclamation points.
	inString = inString.replace(";",r"\;")	# Escape semi-colons points.
	inString = inString.replace('"',r'\"')	# Escape double-quotes.
	inString = inString.replace("`",r"'")	# Replace back-ticks with single quotes.
	return inString

########################################
def updateVar(name, value, folder):
	# (This method isn't currently being used).
	if name not in indigo.variables:
		indigo.variable.create(name, value=value, folder=folder)
	else:
		indigo.variable.updateValue(name, value)

################################################################################
class Plugin(indigo.PluginBase):
	########################################
	# Class Properties
	########################################
	# (none for now).
	
	########################################
	# Plugin Initialization and Destruction.
	########################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs): 
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		self.deviceList = []
		self.debug = pluginPrefs.get("showDebugInfo", False)
		self.debugLog("Initializing Plugin.")
		self.debugLog("Subscribing to all variable changes.")
		# Subscribe to variable changes for all the Announcer devices.
		indigo.variables.subscribeToChanges()
	
	########################################
	def __del__(self):
		self.debugLog("Stopping plugin.")
		indigo.PluginBase.__del__(self)
	
	########################################
	def closedPrefsConfigUi(self, valuesDict, userCancelled):
		if not userCancelled:
			self.debug = valuesDict.get("showDebugInfo", False)
			if self.debug:
				indigo.server.log("Debug logging enabled")
			else:
				indigo.server.log("Debug logging disabled")
				
	
	########################################
	# Action-Specific Methods
	########################################
	def validateActionConfigUi(self, valuesDict, typeId, devId):
		self.debugLog("Starting action config validation for type: " + typeId + ".")
		errorsDict = indigo.Dict()
		anError = False
		
		# Create an appropriate action descriptive text.
		descString = ""
		if typeId == "announce":
			if valuesDict['useMessageVariable'] == False:
				if valuesDict['announceAudioFile'] != "":
					descString = u"play a sound file and speak a message: " + valuesDict['message']
				else:
					descString = u"speak a message: " + valuesDict['message']
			else:
				if valuesDict['announceAudioFile'] != "":
					descString = u"play a sound file and speak a message in variable: " + indigo.variables[int(valuesDict['messageVar'])].name
				else:
					descString = u"speak a message in variable: " + indigo.variables[int(valuesDict['messageVar'])].name
		valuesDict['description'] = descString
		
		# Check the valuesDict for required settings.
		if valuesDict['announceAudioFile'] != "":
			try:
				f = valuesDict['announceAudioFile']
				# Make sure the file exists using the mactypes Alias method.
				f = mactypes.Alias(f)
				self.debugLog("Pre-Speech Audio File set to " + f.path + ".")
				# Set the valuesDict['announceAudioFile'] to the actual path in case it was a link.
				valuesDict['announceAudioFile'] = f.path
				# The file exists, but is it a file or a directory?
				if os.path.isdir(valuesDict['announceAudioFile']) == True:
					errorsDict['announceAudioFile'] = "Cannot play directories. Please specify an audio file."
					self.errorLog(errorsDict['announceAudioFile'])
					anError = True
				else:
					# It's not a directory, now test to see if it's an audio file using the system's "file" command.
					f = subprocess.Popen(["file", valuesDict['announceAudioFile']], stdout=subprocess.PIPE)
					f = str(f.stdout.readline()).lower()
					self.debugLog("System \"file\" examination results: " + f)
					# Split the find output results so we're only examining the file analysis and not the file name.
					f = f.split(":")[1]
					if "audio" in f or "mpeg" in f:
						self.debugLog("File " + valuesDict['announceAudioFile'] + " is an audio file.")
					else:
						errorsDict['announceAudioFile'] = "Not an audio file. Plesae specify an audio file."
						self.errorLog(errorsDict['announceAudioFile'])
						anError = True
			except ValueError, e:
				errorsDict['announceAudioFile'] = "File does not exist. Verify the path is correct."
				self.errorLog(errorsDict['announceAudioFile'])
				self.debugLog("Error returned: " + str(e))
				anError = True
			except Exception, e:
				errorsDict['announceAudioFile'] = "Error with Pre-Speech Audio File value"
				self.errorLog(errorsDict['announceAudioFile'] + ": " + str(e))
				anError = True
		else:
			self.debugLog("Pre-Speech Audio File not being used.")
		
		# Announcement Volume.
		if valuesDict['announceVolume'] != "":
			try:
				# If the integer casting failes, it was an invalid entry.
				f = int(valuesDict['announceVolume'])
				# If the integer casting succeeded, make sure it's a sane number.
				if f < 0 or f > 255:
					errorsDict['announceVolume'] = "Announcement Volume must be between 0 and 255."
					self.errorLog(errorsDict['announceVolume'])
					anError = True
				else:
					self.debugLog("Announcement Volume set to " + str(f) + ".")
			except ValueError:
				errorsDict['announceVolume'] = "Announcement Volume must be a whole number between 0 and 255."
				self.errorLog(errorsDict['announceVolume'])
				anError = True
			except Exception, e:
				errorsDict['announceVolume'] = "Error with Announcement Volume"
				self.errorLog(errorsDict['announceVolume'] + ": " + str(e))
				anError = True
		else:
			# If an announcement audio file was specified, not specifying a volume is an error.
			if valuesDict['announceAudioFile'] != "":
				errorsDict['announceVolume'] = "Announcement Volume must be between 0 and 255 if a Pre-Speech Audio File is specified."
				self.errorLog(errorsDict['announceVolume'])
				anError = True
			else:
				# Since no file was specified, just set the announceVolume to 0 so it has a value.
				valuesDict['announceVolume'] = "0"
		
		# Speech Volume.
		if valuesDict['speechVolume'] != "":
			try:
				# If the integer casting failes, it was an invalid entry.
				f = int(valuesDict['speechVolume'])
				# If the integer casting succeeded, make sure it's a sane number.
				if f < 0 or f > 255:
					errorsDict['speechVolume'] = "Speech Volume must be between 0 and 255."
					self.errorLog(errorsDict['speechVolume'])
					anError = True
				else:
					self.debugLog("Speech Volume set to " + str(f) + ".")
			except ValueError:
				errorsDict['speechVolume'] = "Speech Volume must be a whole number between 0 and 255."
				self.errorLog(errorsDict['speechVolume'])
				anError = True
			except Exception, e:
				errorsDict['speechVolume'] = "Error with Speech Volume"
				self.errorLog(errorsDict['speechVolume'] + ": " + str(e))
				anError = True
		else:
			errorsDict['speechVolume'] = "Speech Volume must be between 0 and 255."
			self.errorLog(errorsDict['speechVolume'])
			anError = True
		
		# Speech Delay.
		if valuesDict['speechDelay'] != "":
			try:
				# If the float casting failes, it was an invalid entry.
				f = float(valuesDict['speechDelay'])
				# If the float casting succeeded, make sure it's a sane number.
				if f < 0 or f > 10:
					errorsDict['speechDelay'] = "Speech Delay must be between 0 and 10."
					self.errorLog(errorsDict['speechDelay'])
					anError = True
				else:
					self.debugLog("Speech Delay set to " + str(f) + ".")
			except ValueError:
				errorsDict['speechDelay'] = "Speech Dealy must be a number between 0 and 10."
				self.errorLog(errorsDict['speechDelay'])
				anError = True
			except Exception, e:
				errorsDict['speechDelay'] = "Error with Speech Delay"
				self.errorLog(errorsDict['speechDelay'] + ": " + str(e))
				anError = True
		else:
			errorsDict['speechDelay'] = "Speech Delay must be between 0 and 10."
			self.errorLog(errorsDict['speechDelay'])
			anError = True
		
		# Speech rate value.
		if valuesDict['rate'] != "":
			try:
				# If the int casting failes, it was an invalid entry.
				f = int(valuesDict['rate'])
				# If the int casting succeeded, make sure it's a sane number.
				if f < 10 or f > 1500:
					errorsDict['rate'] = "Speech Rate must be between 10 and 1500."
					self.errorLog(errorsDict['rate'])
					anError = True
				else:
					self.debugLog("Speech Rate set to " + str(f) + ".")
			except ValueError:
				errorsDict['rate'] = "Speech Rate must be a number between 10 and 1500."
				self.errorLog(errorsDict['rate'])
				anError = True
			except Exception, e:
				errorsDict['rate'] = "Error with Speech Rate"
				self.errorLog(errorsDict['rate'] + ": " + str(e))
				anError = True
		else:
			self.debugLog("Speech Rate not specified. Will use default voice rate.")
		
		# Verify message or variable used to send announcements to the device.
		if valuesDict['useMessageVariable'] == False:
			if valuesDict['message'] != "":
				self.debugLog("Custom message to say: " + valuesDict['message'])
			else:
				errorsDict['message'] = "Please specify a message to announce."
				self.errorLog("No custom message was entered. " + errorsDict['message'])
				anError = True
		else:
			if valuesDict['messageVar'] != "":
				self.debugLog("Speech Text Variable set to " + str(valuesDict['messageVar']) + ".")
			else:
				errorsDict['messageVar'] = "A Speech Message Variable must be selected to continue."
				self.errorLog(errorsDict['messageVar'])
				anError = True
		
		# Make sure a voice is selected.
		if valuesDict['voice'] != "":
			isValid = False
			# Use the AppKit function imported at top to get a list of
			#   installed voices that can be used.
			tempList = NSSpeechSynthesizer.availableVoices()
			self.debugLog("validateDeviceConfigUi: NSSpeechSynthesizer voice list: " + str(tempList))
			# Loop through the list to make the values easily readable.
			for v in tempList:
				# Get the attributres for this voice.
				vAttributes = NSSpeechSynthesizer.attributesForVoice_(v)
				# Get the display name for the voice.
				vName = vAttributes['VoiceName']
				self.debugLog("voice display name: " + vName)
				# Make sure they selected a voice in the list.
				if valuesDict['voice'] == v:
					isValid = True
				v = None
		else:
			errorsDict['voice'] = "A Voice must be selected to continue."
			self.errorLog(errorsDict['voice'])
			anError = True
		
		# Make sure the audioPlayer was specified, regardless of if the useSpecificApp was checked.
		if valuesDict['audioPlayer'] != "":
			# If the useSpecificApp option is false, the only valid value for
			#   audioPlayer is "Play Sound".  Make sure that is the case.
			if valuesDict['useSpecificApp'] == False:
				valuesDict['audioPlayer'] = "Play Sound"
			# Use the appscript "app" method to verify the application exists.
			try:
				f = app(valuesDict['audioPlayer'])
			except ApplicationNotFoundError:
				errorsDict['useSpecificApp'] = "Application \"" + str(valuesDict['audioPlayer']) + "\" could not be found."
				# If the user has elected to specify a path to Play Sound, change the error to make more sense.
				if valuesDict['useSpecificApp'] == True:
					self.errorLog(errorsDict['useSpecificApp'] + " Make sure the application name and path are correct.")
				else:
					self.errorLog(errorsDict['useSpecificApp'] + " If you've changed the name of the application, check the \"Use a Specific Copy of 'Play Sound' App\" check box to specify the application name.")
				anError = True
			except Exception, e:
				errorsDict['useSpecificApp'] = "Unable to validate \"" + str(valuesDict['audioPlayer']) + "\" application."
				self.errorLog(errorsDict['useSpecificApp'] + "  " + str(e))
				anError = True
		else:
			# Make sure the error makes sense based on whether they're trying to specify a location for Play Sound.
			if valuesDict['useSpecificApp'] == True:
				errorsDict['useSpecificApp'] = "No Play Sound application path was specified."
				self.errorLog(errorsDict['useSpecificApp'])
				anError = True
			else:
				# They must have checked the option to specify a path to Play Sound then
				#   deleted the default value before unchecking that option (thus leaving it blank).
				#   Just assign the correct value here.
				valuesDict['audioPlayer'] = "Play Sound"
		
		# Define the "address" value if it isn't defined with the current voice.
		if valuesDict['voice'] not in valuesDict['address']:
			valuesDict['address'] = valuesDict['voice'] + "." + str(random.randint(1,10000))
		
		self.debugLog("Saving values: " + str(valuesDict))
		
		if anError == True:
			return (False, valuesDict, errorsDict)
		else:
			return (True, valuesDict)
	
	########################################
	# Device-Specific Methods
	########################################
	def validateDeviceConfigUi(self, valuesDict, typeId, devId):
		self.debugLog("Starting device config validation.")
		errorsDict = indigo.Dict()
		anError = False
		# Check the valuesDict for required settings.
		if valuesDict['announceAudioFile'] != "":
			try:
				f = valuesDict['announceAudioFile']
				# Make sure the file exists using the mactypes Alias method.
				f = mactypes.Alias(f)
				self.debugLog("Pre-Speech Audio File set to " + f.path + ".")
				# Set the valuesDict['announceAudioFile'] to the actual path in case it was a link.
				valuesDict['announceAudioFile'] = f.path
				# The file exists, but is it a file or a directory?
				if os.path.isdir(valuesDict['announceAudioFile']) == True:
					errorsDict['announceAudioFile'] = "Cannot play directories. Please specify an audio file."
					self.errorLog(errorsDict['announceAudioFile'])
					anError = True
				else:
					# It's not a directory, now test to see if it's an audio file using the system's "file" command.
					f = subprocess.Popen(["file", valuesDict['announceAudioFile']], stdout=subprocess.PIPE)
					f = str(f.stdout.readline()).lower()
					self.debugLog("System \"file\" examination results: " + f)
					# Split the find output results so we're only examining the file analysis and not the file name.
					f = f.split(":")[1]
					if "audio" in f or "mpeg" in f:
						self.debugLog("File " + valuesDict['announceAudioFile'] + " is an audio file.")
					else:
						errorsDict['announceAudioFile'] = "Not an audio file. Plesae specify an audio file."
						self.errorLog(errorsDict['announceAudioFile'])
						anError = True
			except ValueError, e:
				errorsDict['announceAudioFile'] = "File does not exist. Verify the path is correct."
				self.errorLog(errorsDict['announceAudioFile'])
				self.debugLog("Error returned: " + str(e))
				anError = True
			except Exception, e:
				errorsDict['announceAudioFile'] = "Error with Pre-Speech Audio File value"
				self.errorLog(errorsDict['announceAudioFile'] + ": " + str(e))
				anError = True
		else:
			self.debugLog("Pre-Speech Audio File not being used.")
		
		# Announcement Volume.
		if valuesDict['announceVolume'] != "":
			try:
				# If the integer casting failes, it was an invalid entry.
				f = int(valuesDict['announceVolume'])
				# If the integer casting succeeded, make sure it's a sane number.
				if f < 0 or f > 255:
					errorsDict['announceVolume'] = "Announcement Volume must be between 0 and 255."
					self.errorLog(errorsDict['announceVolume'])
					anError = True
				else:
					self.debugLog("Announcement Volume set to " + str(f) + ".")
			except ValueError:
				errorsDict['announceVolume'] = "Announcement Volume must be a whole number between 0 and 255."
				self.errorLog(errorsDict['announceVolume'])
				anError = True
			except Exception, e:
				errorsDict['announceVolume'] = "Error with Announcement Volume"
				self.errorLog(errorsDict['announceVolume'] + ": " + str(e))
				anError = True
		else:
			# If an announcement audio file was specified, not specifying a volume is an error.
			if valuesDict['announceAudioFile'] != "":
				errorsDict['announceVolume'] = "Announcement Volume must be between 0 and 255 if a Pre-Speech Audio File is specified."
				self.errorLog(errorsDict['announceVolume'])
				anError = True
			else:
				# Since no file was specified, just set the announceVolume to 0 so it has a value.
				valuesDict['announceVolume'] = "0"
		
		# Speech Volume.
		if valuesDict['speechVolume'] != "":
			try:
				# If the integer casting failes, it was an invalid entry.
				f = int(valuesDict['speechVolume'])
				# If the integer casting succeeded, make sure it's a sane number.
				if f < 0 or f > 255:
					errorsDict['speechVolume'] = "Speech Volume must be between 0 and 255."
					self.errorLog(errorsDict['speechVolume'])
					anError = True
				else:
					self.debugLog("Speech Volume set to " + str(f) + ".")
			except ValueError:
				errorsDict['speechVolume'] = "Speech Volume must be a whole number between 0 and 255."
				self.errorLog(errorsDict['speechVolume'])
				anError = True
			except Exception, e:
				errorsDict['speechVolume'] = "Error with Speech Volume"
				self.errorLog(errorsDict['speechVolume'] + ": " + str(e))
				anError = True
		else:
			errorsDict['speechVolume'] = "Speech Volume must be between 0 and 255."
			self.errorLog(errorsDict['speechVolume'])
			anError = True
		
		# Speech Delay.
		if valuesDict['speechDelay'] != "":
			try:
				# If the float casting failes, it was an invalid entry.
				f = float(valuesDict['speechDelay'])
				# If the float casting succeeded, make sure it's a sane number.
				if f < 0 or f > 10:
					errorsDict['speechDelay'] = "Speech Delay must be between 0 and 10."
					self.errorLog(errorsDict['speechDelay'])
					anError = True
				else:
					self.debugLog("Speech Delay set to " + str(f) + ".")
			except ValueError:
				errorsDict['speechDelay'] = "Speech Dealy must be a number between 0 and 10."
				self.errorLog(errorsDict['speechDelay'])
				anError = True
			except Exception, e:
				errorsDict['speechDelay'] = "Error with Speech Delay"
				self.errorLog(errorsDict['speechDelay'] + ": " + str(e))
				anError = True
		else:
			errorsDict['speechDelay'] = "Speech Delay must be between 0 and 10."
			self.errorLog(errorsDict['speechDelay'])
			anError = True
		
		# Speech rate value.
		if valuesDict['rate'] != "":
			try:
				# If the int casting failes, it was an invalid entry.
				f = int(valuesDict['rate'])
				# If the int casting succeeded, make sure it's a sane number.
				if f < 10 or f > 1500:
					errorsDict['rate'] = "Speech Rate must be between 10 and 1500."
					self.errorLog(errorsDict['rate'])
					anError = True
				else:
					self.debugLog("Speech Rate set to " + str(f) + ".")
			except ValueError:
				errorsDict['rate'] = "Speech Rate must be a number between 10 and 1500."
				self.errorLog(errorsDict['rate'])
				anError = True
			except Exception, e:
				errorsDict['rate'] = "Error with Speech Rate"
				self.errorLog(errorsDict['rate'] + ": " + str(e))
				anError = True
		else:
			self.debugLog("Speech Rate not specified. Will use default voice rate.")
		
		# Verify variable used to send announcements to the device.
		if valuesDict['messageVar'] != "":
			self.debugLog("Speech Text Variable set to " + str(valuesDict['messageVar']) + ".")
		else:
			errorsDict['messageVar'] = "A Speech Text Variable must be selected to continue."
			self.errorLog(errorsDict['messageVar'])
			anError = True
		
		# Make sure a voice is selected.
		if valuesDict['voice'] != "":
			isValid = False
			# Use the AppKit function imported at top to get a list of
			#   installed voices that can be used.
			tempList = NSSpeechSynthesizer.availableVoices()
			self.debugLog("validateDeviceConfigUi: NSSpeechSynthesizer voice list: " + str(tempList))
			# Loop through the list to make the values easily readable.
			for v in tempList:
				# Get the attributres for this voice.
				vAttributes = NSSpeechSynthesizer.attributesForVoice_(v)
				# Get the display name for the voice.
				vName = vAttributes['VoiceName']
				self.debugLog("voice display name: " + vName)
				# Make sure they selected a voice in the list.
				if valuesDict['voice'] == v:
					isValid = True
				v = None
		else:
			errorsDict['voice'] = "A Voice must be selected to continue."
			self.errorLog(errorsDict['voice'])
			anError = True
		
		# Make sure the audioPlayer was specified, regardless of if the useSpecificApp was checked.
		if valuesDict['audioPlayer'] != "":
			# If the useSpecificApp option is false, the only valid value for
			#   audioPlayer is "Play Sound".  Make sure that is the case.
			if valuesDict['useSpecificApp'] == False:
				valuesDict['audioPlayer'] = "Play Sound"
			# Use the appscript "app" method to verify the application exists.
			try:
				f = app(valuesDict['audioPlayer'])
			except ApplicationNotFoundError:
				errorsDict['useSpecificApp'] = "Application \"" + str(valuesDict['audioPlayer']) + "\" could not be found."
				# If the user has elected to specify a path to Play Sound, change the error to make more sense.
				if valuesDict['useSpecificApp'] == True:
					self.errorLog(errorsDict['useSpecificApp'] + " Make sure the application name and path are correct.")
				else:
					self.errorLog(errorsDict['useSpecificApp'] + " If you've changed the name of the application, check the \"Use a Specific Copy of 'Play Sound' App\" check box to specify the application name.")
				anError = True
			except Exception, e:
				errorsDict['useSpecificApp'] = "Unable to validate \"" + str(valuesDict['audioPlayer']) + "\" application."
				self.errorLog(errorsDict['useSpecificApp'] + "  " + str(e))
				anError = True
		else:
			# Make sure the error makes sense based on whether they're trying to specify a location for Play Sound.
			if valuesDict['useSpecificApp'] == True:
				errorsDict['useSpecificApp'] = "No Play Sound application path was specified."
				self.errorLog(errorsDict['useSpecificApp'])
				anError = True
			else:
				# They must have checked the option to specify a path to Play Sound then
				#   deleted the default value before unchecking that option (thus leaving it blank).
				#   Just assign the correct value here.
				valuesDict['audioPlayer'] = "Play Sound"
		
		# Define the "address" value if it isn't defined with the current voice.
		if valuesDict['voice'] not in valuesDict['address']:
			valuesDict['address'] = valuesDict['voice'] + "." + str(random.randint(1,10000))
		
		self.debugLog("Saving values: " + str(valuesDict))
		
		if anError == True:
			return (False, valuesDict, errorsDict)
		else:
			return (True, valuesDict)
	
	########################################
	def updateDeviceState(self, device, state, newValue):
		self.debugLog("Updating device state.")
		if (newValue != device.states[state]):
			device.updateStateOnServer(key=state, value=newValue)
	
	########################################
	def deviceStartComm(self, dev):
		self.debugLog("Starting device: " + dev.name)
		if dev.id not in self.deviceList:
			self.deviceList.append(dev.id)
			self.updateDeviceState(dev, "status", "idle")
	
	########################################
	def deviceStopComm(self, dev):
		self.debugLog("Stopping device: " + dev.name)
		if dev.id in self.deviceList:
			self.deviceList.remove(dev.id)
			self.updateDeviceState(dev, "status", "idle")
	
	########################################
	def listVoices(self, filter="", valuesDict=None, typeId="", targetId=0):
		# Get the list of installed voices on the system.
		voiceList = []
		
		# Use the AppKit function imported at top to get a list of
		#   installed voices that can be used.
		tempList = NSSpeechSynthesizer.availableVoices()
		self.debugLog("NSSpeechSynthesizer voice list: " + str(tempList))
		# Loop through the list to make the values easily readable.
		for v in tempList:
			# Get the attributres for this voice.
			vAttributes = NSSpeechSynthesizer.attributesForVoice_(v)
			# Get the display name for the voice.
			vName = vAttributes['VoiceName']
			self.debugLog("voice display name: " + vName)
			# Make it into a touple for menu item use.
			vName = (vName, vName)
			# Add it to the array.
			voiceList.append(vName)
			# Clear the values.
			vName = None
		# Send back the voice list.
		return voiceList
	
	########################################
	# Main Announcement Method
	########################################
	def announce(self, dev):
		self.debugLog("Running announce method. Data passed to the method: " + str(dev))
		
		if dev.deviceId == 0:
			# This is an Action request, not a Device update action.
			
			# Make sure all the necessary device properties exist
			#   by assigning them to local variables.
			try:
				audioPlayer = dev.props["audioPlayer"]
				announceAudioFile = dev.props["announceAudioFile"]
				announceVolume = int(dev.props["announceVolume"])	# integer
				speechDelay = float(dev.props["speechDelay"])		# real
				rate = dev.props["rate"]
				messageVar = dev.props["messageVar"]
				message = dev.props["message"]
				useMessageVariable = dev.props["useMessageVariable"]
				messageFile = dev.props["messageFile"]
				voice = dev.props["voice"]
				speechVolume = dev.props["speechVolume"]
				self.debugLog("action property values assigned.")
			except Exception, e:
				
				self.errorLog("Unable to get action properties: " + str(e))
				# We can't really recover from this error.
				return
			
			# Use the "app" appscript method to define and launch the audio player.
			try:
				audioPlayer = app(audioPlayer)
				# If the above assignment doesn't throw an error, the audio player is running.
				self.debugLog("Audio player " + str(dev.props["audioPlayer"]) + " is running.")
			except:
				# If the above audioPlay assignment threw an exception, the audio player probably isn't running.
				self.errorLog("Audio player " + str(dev.props["audioPlayer"]) + " does not appear to be installed.")
				# We can't recover from this error.
				return
			
			if useMessageVariable == True:
				# If we're not using a custom message, get the content of the variable to be read.
				try:
					varText = indigo.variables[int(messageVar)].value
				except Exception, e:
					self.errorLog("Unable to get the content of the specified variable (ID: " + messageVar + ").")
					return
				# Make the variable content the message to speak.
				message = varText
				self.debugLog("Got variable " + indigo.variables[int(messageVar)].name + " contents to speak. Contents: " + message)
				
			# If the message contains text, continue. Otherwise, don't do anything.
			if message != "":
				# Define the name of the spoken message file based on the time.
				messageFile = u"/tmp/" + str(time.time()) + ".aif"
				
				# If an announcement audio file was specified, verify it exists.
				if announceAudioFile != "":
					# Use the Alias method of the mactypes libraries to convert the UNIX path of the
					#   announcement audio file to a Mac HFS path. This also servces to verify that the file exists.
					try:
						self.debugLog("Converting announcement audio file UNIX path to Mac HFS filesystem path.")
						announceAudioFile = mactypes.Alias(announceAudioFile).hfspath
					except Exception, e:
						self.errorLog("Error converting announcement audio file UNIX pat hto Mac HFS filesystem path: " + str(e))
				
				#############################
				# Create Spoken Announcement
				#############################
				
				# Use the "say" command to create an AIFF audio file with the requested spoken message.
				indigo.server.log("Announcing message \"" + message + "\".", "Announcer")
				self.debugLog("Using \"say\" command to create message file \"" + messageFile + "\" with spoken message \"" + message + "\" using voice " + voice + " at volume " + speechVolume + ".")
				
				# It doesn't appear to be necessary to escape special characters when using the "call"
				#   method unless using the "shell=True" parameter (which we aren't). So we'll just give
				#   the "say" command the whole text of the message.
				
				# If the rate is empty, this indicates that a rate should not be specified.
				if rate != "":
					sayResult = subprocess.check_call(["say", "-v", voice, "-r", rate, "-o", messageFile, message])
				else:
					sayResult = subprocess.check_call(["say", "-v", voice, "-o", messageFile, message])
				
				self.debugLog("Say command results: " + str(sayResult))
				
				# If the result of the "say" command was not zero, there was an error.
				if sayResult != 0:
					self.errorLog("There was an error while trying to create the spoken text audio file. Please enable debugging in the Announcer plugin configuration and try again for more information.")
					return
				
				# Only play the announcement audio file if it was specified.
				if announceAudioFile != "":
					# Tell the audio player (Play Sound) to play the announcement audio file first.
					self.debugLog("Telling " + str(dev.props["audioPlayer"]) + " to play the file " + str(dev.props["announceAudioFile"]) + ".")
					audioPlayer.play(announceAudioFile, concurrence=True, at_volume=announceVolume)
				
				# Wait the specified amount of time before playing the spoken message.
				self.sleep(speechDelay)
				
				# Now play the spoken text audio file.
				self.debugLog("Converting message audio file UNIX path to Mac HFS filesystem path.")
				messageFile = mactypes.Alias(messageFile).hfspath	# Convert from UNIX to HFS path.
				self.debugLog("Telling " + str(dev.props["audioPlayer"]) + " to play the file " + messageFile + ".")
				audioPlayer.play(messageFile, concurrence=True, at_volume=speechVolume)
			else:
				indigo.server.log("No text was provided. Nothing will be announced.", "Announcer")
				
		else:
			# This is a Device update request.
			
			# Set the status.
			self.updateDeviceState(dev, "status", "idle")
			
			localProps = dev.pluginProps	# Local copy for manipulation.
			# Make sure all the necessary device properties exist
			#   by assigning them to local variables.
			try:
				audioPlayer = dev.pluginProps["audioPlayer"]
				announceAudioFile = dev.pluginProps["announceAudioFile"]
				announceVolume = int(dev.pluginProps["announceVolume"])	# integer
				speechDelay = float(dev.pluginProps["speechDelay"])		# real
				rate = dev.pluginProps["rate"]
				messageVar = dev.pluginProps["messageVar"]
				message = dev.pluginProps["message"]
				messageFile = dev.pluginProps["messageFile"]
				voice = dev.pluginProps["voice"]
				speechVolume = dev.pluginProps["speechVolume"]
				self.debugLog("pluginProps values assigned.")
			except Exception, e:
				# Set the status.
				self.updateDeviceState(dev, "status", "error")
				
				self.errorLog("Unable to get device properties: " + str(e))
				# We can't necessarily recover from this error.
				return
			
			# Use the "app" appscript method to define and launch the audio player.
			try:
				audioPlayer = app(audioPlayer)
				# If the above assignment doesn't throw an error, the audio player is running.
				self.debugLog("Audio player " + str(dev.pluginProps["audioPlayer"]) + " is running.")
			except:
				# Set the status.
				self.updateDeviceState(dev, "status", "error")
				
				# If the above audioPlay assignment threw an exception, the audio player probably isn't running.
				self.errorLog("Audio player " + str(dev.pluginProps["audioPlayer"]) + " does not appear to be installed.")
				# We can't recover from this error.
				return
			
			# Don't do anything if the message content is blank.
			if message != "":
				# Define the name of the spoken message file based on the time.
				messageFile = u"/tmp/" + str(time.time()) + ".aif"
				localProps["messageFile"] = messageFile
				# Update the device properties.
				dev.replacePluginPropsOnServer(localProps)
				
				# If an announcement audio file was specified, verify it exists.
				if announceAudioFile != "":
					# Use the Alias method of the mactypes libraries to convert the UNIX path of the
					#   announcement audio file to a Mac HFS path. This also servces to verify that the file exists.
					try:
						self.debugLog("Converting announcement audio file UNIX path to Mac HFS filesystem path.")
						announceAudioFile = mactypes.Alias(announceAudioFile).hfspath
					except Exception, e:
						self.errorLog("Error converting announcement audio file UNIX pat hto Mac HFS filesystem path: " + str(e))
				
				#############################
				# Create Spoken Announcement
				#############################
				# Set the status.
				self.updateDeviceState(dev, "status", "working")
				
				# Use the "say" command to create an AIFF audio file with the requested spoken message.
				indigo.server.log("Announcing message \"" + message + "\".", dev.name)
				self.debugLog("Using \"say\" command to create message file \"" + messageFile + "\" with spoken message \"" + message + "\" using voice " + voice + ".")
				
				# It doesn't appear to be necessary to escape special characters when using the "call"
				#   method unless using the "shell=True" parameter (which we aren't). So we'll just give
				#   the "say" command the whole text of the message.
				
				# If the rate is empty, this indicates that a rate should not be specified.
				if rate != "":
					subprocess.call(["say", "-v", voice, "-r", rate, "-o", messageFile, message])
				else:
					subprocess.call(["say", "-v", voice, "-o", messageFile, message])
				
				# Only play the announcement audio file if it was specified.
				if announceAudioFile != "":
					# Tell the audio player (Play Sound) to play the announcement audio file first.
					self.debugLog("Telling " + str(dev.pluginProps["audioPlayer"]) + " to play the file " + str(dev.pluginProps["announceAudioFile"]) + ".")
					audioPlayer.play(announceAudioFile, concurrence=True, at_volume=announceVolume)
				
				# Set the status.
				self.updateDeviceState(dev, "status", "waiting")
				
				# Wait the specified amount of time before playing the spoken message.
				self.sleep(speechDelay)
				
				# Set the status.
				self.updateDeviceState(dev, "status", "working")
				
				# Now play the spoken text audio file.
				self.debugLog("Converting message audio file UNIX path to Mac HFS filesystem path.")
				messageFile = mactypes.Alias(messageFile).hfspath	# Convert from UNIX to HFS path.
				self.debugLog("Telling " + str(dev.pluginProps["audioPlayer"]) + " to play the file " + str(dev.pluginProps["messageFile"]) + ".")
				audioPlayer.play(messageFile, concurrence=True, at_volume=speechVolume)
			else:
				indigo.server.log("Variable contains no text. Nothing will be announced.", dev.name)
			
			# Set the status.
			self.updateDeviceState(dev, "status", "idle")
		
	########################################
	# Variable Change Functions
	########################################
	def variableCreated(self, var):
		indigo.PluginBase.variableCreated(self, var)
		self.debugLog("Variable " + str(var.name) + " created.")
	
	def variableDeleted(self, var):
		indigo.PluginBase.variableDeleted(self, var)
		self.debugLog("Variable " + str(var.name) + " deleted.")
		# Find any Announcer device that monitors this variable.
		for deviceId in self.deviceList:
			# Get the device object.
			device = indigo.devices[deviceId]
			# If this device is monitoring this variable, log an error
			#   Indicating that the device will no longer announce messages.
			if device.pluginProps["messageVar"] == str(var.id):
				self.debugLog("variableUpdated: Device \"" + str(device.name) + "\" was monitoring variable \"" + str(var.name) + "\" which was just deleted.")
				self.errorLog("WARNING: Device \"" + str(device.name) + "\" was using Indigo variable \"" + str(var.name) + "\" for spoken announcements.  That variable was just deleted. Open the \"" + str(device.name) + "\" device configuration dialog to select a new variable to use for spoken announcements.")
	
	def variableUpdated(self, origVar, newVar):
		indigo.PluginBase.variableUpdated(self, origVar, newVar)
		newValue = newVar.value
		# Make sure the new value is different from the old one.
		if origVar.value != newVar.value:
			# Find any Announcer device that monitors this variable.
			for deviceId in self.deviceList:
				# Get the device object.
				device = indigo.devices[deviceId]
				localProps = device.pluginProps
				# If this device is monitoring this variable, update the
				#   device pluginProps and play the text of the new value.
				if device.pluginProps["messageVar"] == str(newVar.id):
					self.debugLog("variableUpdated: Device \"" + str(device.name) + "\" is monitoring variable \"" + str(newVar.name) + "\"... Updating device properties and calling announce method.")
					localProps["message"] = str(newValue)
					device.replacePluginPropsOnServer(localProps)
					
					self.announce(device)
	
